{
  This demo project shows how to use regular expressions for searching in
  TRichView.

  This is a simple version of this demo: searching from cursor

  This demo requires PCRE Wrapper
  http://renatomancuso.com/software/dpcre/dpcre.htm
  Modified:
  https://github.com/rubiot/ibiblia/blob/master/pcre_dll.pas (and also PCRE.pas
  from the same location)
  ---

  https://www.trichview.com

}


unit MainFrm;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, SysUtils, Classes, Graphics,
  Controls, Forms, Dialogs, Buttons, ExtCtrls,
  RichView, RVEdit, StdCtrls, RVStyle, PCRE, RVTypes;

type
  TfrmMain = class(TForm)
    rve: TRichViewEdit;
    Panel1: TPanel;
    btnOpen: TSpeedButton;
    btnFind: TSpeedButton;
    Label1: TLabel;
    txtRegEx: TEdit;
    cbIgnoreCase: TCheckBox;
    RVStyle1: TRVStyle;
    OpenDialog1: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure rveChange(Sender: TObject);
    procedure btnFindClick(Sender: TObject);
    procedure txtRegExChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbIgnoreCaseClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
  private
    { Private declarations }
    FTextChanged, FExpressionChanged: Boolean;
    FText: String;
    FTextW: TRVUnicodeString;
    FRegEx: IRegex;
    procedure UpdateText;
    procedure Select(const M: IMatch);
    function GetSearchOptions: TRegCompileOptions;
  public
    { Public declarations }
    procedure Search(StartPos: Integer);
  end;

var
  frmMain: TfrmMain;

implementation

uses
  RVLinear, CRVData, CRVFData, RVStrFuncs, RVUni;

{$R *.lfm}

{ TfrmMain }


procedure TfrmMain.FormCreate(Sender: TObject);
begin
  FTextChanged := True;
  FExpressionChanged := True;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  FRegEx := nil;
end;

// Converting length of substring of UTF-16 string to length of UTF-8 string
// The substring is defined as a substring of S, starting from charactter FromIndex,
// having length CharCount
// The result is correct only if this substring contains all its surrogate pairs
// completely. Correctness of S is not checked.
function UTF16ToUTF8CharCount(const S: TRVUnicodeString; FromIndex, CharCount: Integer): Integer;
var
  i: Integer;
begin
  Result := 0;
  if CharCount = 0 then
    exit;
  if RVGet1Index(FromIndex) + CharCount - 1 > Length(S) then
    CharCount := Length(S) - RVGet1Index(FromIndex) + 1;
  for i := FromIndex to FromIndex + CharCount - 1 do
    case ord(S[i]) of
      $0000..$007F:
        inc(Result);
      $0080..$07FF:
        inc(Result, 2);
      $0800..$D7FF, $E000..$FFFF:
        inc(Result, 3);
      $D800..$DBFF: // high surrogates
        inc(Result, 4);
      $DC00..$DFFF: // low surrogates
        ;
    end;
end;

// Converting length of substring of UTF-8 string to length of UTF-16 string
// The substring is defined as a substring of S, starting from charactter FromIndex,
// having length CharCount
// The result is correct only if this substring contains all UTF-8 characters
// completely. Correctness of S is not checked.
function UTF8ToUTF16CharCount(const S: TRVUTF8String; FromIndex, CharCount: Integer): Integer;
var
  i: Integer;
  Ch: Byte;
begin
  Result := 0;
  if CharCount = 0 then
    exit;
  if RVGet1Index(FromIndex) + CharCount - 1 > Length(S) then
    CharCount := Length(S) - RVGet1Index(FromIndex) + 1;
  for i := FromIndex to FromIndex + CharCount - 1 do
  begin
    Ch := ord(S[i]);
    if ((Ch and $80) = 0) or ((Ch and $E0) = $C0) or ((Ch and $F0) = $E0) then
      inc(Result)
    else if ((Ch and $F8) = $F0) then
      inc(Result, 2);
  end;
end;

procedure TfrmMain.btnOpenClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    rve.Clear;
    rve.DeleteUnusedStyles(True, True, True);
    case OpenDialog1.FilterIndex of
      1: rve.LoadRVF(UTF8Decode(OpenDialog1.FileName));
      2: rve.LoadRTF(UTF8Decode(OpenDialog1.FileName));
      3: rve.LoadText(UTF8Decode(OpenDialog1.FileName), 0, 0, False);
      4: rve.LoadTextW(UTF8Decode(OpenDialog1.FileName), 0, 0, False);
    end;
    rve.Format;
    rveChange(Sender);
  end;
end;

procedure TfrmMain.rveChange(Sender: TObject);
begin
  FTextChanged := True;
  txtRegExChange(Sender);
end;

procedure TfrmMain.txtRegExChange(Sender: TObject);
begin
  FExpressionChanged := True;
end;

procedure TfrmMain.cbIgnoreCaseClick(Sender: TObject);
begin
  txtRegExChange(Sender);
end;

procedure TfrmMain.Select(const M: IMatch);
var
  StartIndex, EndIndex, ItemNo1, Offs1, ItemNo2, Offs2: Integer;
  RVData1, RVData2: TCustomRVData;
begin
  StartIndex := UTF8ToUTF16CharCount(FText, 1, M.GetIndex);
  EndIndex := StartIndex + UTF8ToUTF16CharCount(FText, 1 + M.GetIndex, M.GetLength);

  if LinearToRichView(rve, rve.RVData, StartIndex, RVData1, ItemNo1, Offs1) and
     LinearToRichView(rve, rve.RVData, EndIndex, RVData2, ItemNo2, Offs2) and
     (RVData1 = RVData2) then
  begin
    RVData1 := RVData1.Edit;
    TCustomRVFormattedData(RVData1).SetSelectionBounds(ItemNo1, Offs1, ItemNo2, Offs2);
    rve.RefreshAll;
  end
  else
    Application.MessageBox('The matched fragment cannot be selected', 'Error', MB_OK);
end;

procedure TfrmMain.Search(StartPos: Integer);
var
  M: IMatch;
begin
  if FExpressionChanged then
    FRegEx := RegexCreate(txtRegEx.Text, GetSearchOptions);
  M := FRegEx.Match(FText, StartPos);
  if (M = nil) or (M.GetIndex < 0) or (M.GetLength = 0) then
  begin
    Application.MessageBox('The expression does not match', 'End of Search', MB_OK);
    exit;
  end;
  Select(M);
end;


procedure TfrmMain.btnFindClick(Sender: TObject);
var
  StartPos: Integer;
begin
  if txtRegEx.Text = '' then
  begin
    Application.MessageBox('Please enter an expression to search', 'Empty Expression', MB_OK);
    txtRegEx.SetFocus;
    exit;
  end;
  rve.SetFocus;
  UpdateText;
  StartPos := UTF16ToUTF8CharCount(FTextW, 1, RVGetLinearCaretPos(rve));
  if Length(FText) <= StartPos then
  begin
    Application.MessageBox('End of text', 'End of Search', MB_OK);
    exit;
  end;
  Search(StartPos);
  FExpressionChanged := False;
end;


procedure TfrmMain.UpdateText;
begin
  if FTextChanged then
  begin
    FTextW := RVGetTextRange(rve, 0, RVGetTextLength(rve));
    FText := UTF8Encode(FTextW);
    FTextChanged := False;
  end;
end;

function TfrmMain.GetSearchOptions: TRegCompileOptions;
begin
  Result := [rcoMultiLine, rcoUTF8];
  if cbIgnoreCase.Checked then
    Include(Result, rcoIgnoreCase);
end;

end.
